#!/usr/bin/perl -T
#
# darcs.cgi - the darcs repository viewer
#
# Copyright (c) 2004 Will Glozer
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

#
# This program calls darcs (or its own subroutines) to generate XML
# which is rendered into HTML by XSLT.  It is capable of displaying
# the files in a repository, various patch histories, annotations, etc.
#

use strict;

use CGI qw( :standard );
use CGI::Util;
use File::Basename;
use File::stat;
use File::Temp;
use IO::Handle;
use POSIX;

## the following variables can be customized to reflect your system
## configuration by defining them appropriately in the file
## "/etc/cgi.conf".  The syntax accepts equals signs or simply
## blanks separating values from assignments.

$ENV{'PATH'} = read_conf('PATH', $ENV{'PATH'});

# path to executables, or just the executable if they are in $ENV{'PATH'}
my $darcs_program    = read_conf("darcs", "darcs");
my $xslt_program     = read_conf("xsltproc", "xsltproc");

# directory containing repositories
my $repository_root  = read_conf("reposdir", "/var/www");

# XSLT template locations
my $template_root = read_conf("xslt_dir", "/usr/share/darcs/xslt");

my $xslt_annotate = "$template_root/annotate.xslt";
my $xslt_browse   = "$template_root/browse.xslt";
my $xslt_patches  = "$template_root/patches.xslt";
my $xslt_repos    = "$template_root/repos.xslt";
my $xslt_rss      = "$template_root/rss.xslt";

my $xslt_errors   = "$template_root/errors.xslt";

# CSS stylesheet that XSLT templates refer to.  This is a HTTP request
# path, not a local file system path. The default will caues darcs.cgi
# to serve the stylesheet rather than the web server.
my $stylesheet = read_conf("stylesheet", "/cgi-bin/darcs.cgi/styles.css");

# location of the CSS stylesheet that darcs.cgi will serve if it
# receives a request for '/styles.css'
my $css_styles = read_conf("css_styles", "/etc/darcs/styles.css");

# XML source for the error pages
my $xml_errors = "$template_root/errors.xml";

# encoding to include in XML declaration
my $xml_encoding = read_conf("xml_encoding", "UTF-8");

## end customization

# ----------------------------------------------------------------------

# read a value from the cgi.conf file.

sub read_conf {
  my ($flag, $val);
  if ($#_ == 0) {
    ($flag) = @_;
    $val = "";
  } else {
    ($flag, $val) = @_;
  }
  if (open(CGI_CONF, "/etc/darcs/cgi.conf")) {
    while (<CGI_CONF>) {
      if (/^\s*$flag\s*=\s*(\S+)/) {
        $val = $1;
      } elsif (/^\s*$flag\s+(\S+)/) {
        $val = $1;
      }
    }
    close(CGI_CONF);
  }
  return $val;
}

# transform and output `xml' with stylesheet file `xslt'
sub transform {
    my ($xml, $xslt, $args) = @_;

    print "Content-type: text/html\r\n\r\n";
    my $pipe;
    open $pipe, "| $xslt_program $args $xslt -";

    seek ($xml, 0, 0);
    while (<$xml>) {
      print $pipe $_;
    }
}

# create a temporary XML document with a root element and the repository path
sub make_xml {
    my ($repo, $dir, $file) = @_;
    my $fh = File::Temp::tempfile();
    my ($full_path, $path) = '/';

    printf $fh qq(<?xml version="1.0" encoding="$xml_encoding"?>\n);

    printf $fh qq(<darcs repository="$repo" target="%s/%s%s">\n),
        $repo, ($dir ? "$dir/" : ''), ($file ? "$file" : '');

    print $fh qq(<path>\n);
    foreach $path (split('/', "$repo/$dir")) {
        $full_path .= "$path/";
        print $fh qq(<directory full-path="$full_path">$path</directory>\n);
    }
    if ($file) {
        print $fh qq(<file full-path="$full_path$file">$file</file>\n) if $file;
    }
    print $fh qq(</path>\n\n);

    return $fh;
}

# finish XML output
sub finish_xml {
    my ($fh) = @_;
    print $fh "\n</darcs>\n";
    $fh->flush;
}

# run darcs and collect the output in an XML document
sub darcs_xml {
    my ($repo, $cmd, $args, $dir, $file) = @_;
    my $fh = make_xml($repo, $dir, $file);

    $args .= " --xml-output ";
    darcs($fh, $repo, $cmd, $args, $dir, $file);

    finish_xml($fh);
    return $fh;
}

# run darcs with output redirected to the specified file handle
sub darcs {
    my ($fh, $repo, $cmd, $args, $dir, $file) = @_;
    my ($target);

    # quote target only if there is one, otherwise darcs will get an empty param
    if ($dir || $file) {
        $target = sprintf("%s%s%s", $dir, ($dir ? '/' : ''), $file);
    }

    chdir "$repository_root/$repo";
    open(DARCS, "$darcs_program '$cmd' $args $target|");
    while (<DARCS>) {
        print $fh $_;
    }
    close(DARCS);
}

# get a directory listing as XML output
sub dir_listing {
    my ($repo, $dir) = @_;
    my $fh = make_xml($repo, $dir, '');
    my $file;

    print $fh "<files>\n";
    foreach $file (glob("$repository_root/$repo/_darcs/current/$dir/*")) {
        my $secs  = stat($file)->mtime;
        my $mtime = localtime($secs);
        my $ts = POSIX::strftime("%Y%m%d%H%M%S", gmtime $secs);

        my ($name, $type);

         if (-d $file) {
             ($name, $type) = (basename($file) . '/', 'directory');
         } else {
             ($name, $type) = (basename($file), 'file');
         }
         printf $fh qq(  <$type name="$name" modified="$mtime" ts="$ts" />\n);
    }
    print $fh "</files>\n";

    finish_xml($fh);
    return $fh;
}

# get a repository listing as XML output
sub repo_listing {
    my $fh = make_xml("", "", "");
    my ($repo, $name);

    print $fh "<repositories>\n";
    foreach $repo (glob("$repository_root/*")) {
        if (-d "$repo/_darcs") {
            $name = basename($repo);
            printf $fh qq(  <repository name="$name" />\n);
        }
    }
    print $fh "</repositories>\n";

    finish_xml($fh);
    return $fh;
}

# show an error page
sub show_error {
    my ($type, $code, $message) = @_;
    my $xml;

    # set the xslt processing arguments
    my $xslt_args = qq {
        --stringparam error-type '$type'
        --stringparam stylesheet '$stylesheet'
    };
    $xslt_args =~ s/\s+/ /gm;

    print "Status: $code $message\r\n";

    open($xml, $xml_errors);
    transform($xml, $xslt_errors, $xslt_args);
    close($xml);
}

# safely extract a parameter from the http request.  This applies a regexp
# to the parameter which should group only the appropriate parameter value
sub safe_param {
    my ($param, $regex, $default) = @_;
    my $value = CGI::Util::unescape(param($param));
    return ($value =~ $regex) ? $1 : $default;
}

# common regular expressions for validating passed parameters
my $hash_regex = /^([\w\-.]+)$/;
my $path_regex = qr@^([^\\!\$\^&*()\[\]{}<>~`|';"?\r\n]+)$@;

# respond to a CGI request
sub respond {
    # untaint the full URL to this CGI
    my $cgi_url = CGI::Util::unescape(url());
    $cgi_url =~ $path_regex or die qq(bad url "$cgi_url");
    $cgi_url = $1;

    # untaint script_name, reasonable to expect only \w, -, /, and . in the name
    my $script_name = CGI::Util::unescape(script_name());
    $script_name =~ qr~^([\w/.\-]+)$~ or die qq(bad script_name "$script_name");
    $script_name = $1;

    # untaint simple parameters, which can only have chars matching \w+
    my $cmd  = safe_param('c', '^(\w+)$', 'browse');
    my $sort = safe_param('s', '^(\w+)$', '');

    # set the xslt processing arguments
    my $xslt_args = qq {
        --stringparam cgi-program '$script_name'
        --stringparam cgi-url '$cgi_url'
        --stringparam sort-by '$sort'
        --stringparam stylesheet '$stylesheet'
    };
    $xslt_args =~ s/\s+/ /gm;

    my ($path) = CGI::Util::unescape(path_info());
    # don't allow ./ or ../ in paths
    $path =~ s|[.]+/||g;

    # check whether we're asking for styles.css
    if ($path eq '/styles.css') {
        open (STYLES_CSS, $css_styles) or die qq(couldn't open "${css_styles}");
        print "Content-type: text/css\r\n\r\n";
        while (<STYLES_CSS>) {
          print $_;
        }
        close (STYLES_CSS);
        return;
    }

    # when no repository is requested display available repositories
    if (length($path) < 2) {
        my $fh = repo_listing();
        transform($fh, $xslt_repos, $xslt_args);
        return;
    }

    # don't allow any shell meta characters in paths
    $path =~ $path_regex or die qq(bad path_info "$path");
    my @path = split('/', substr($1, 1));

    # split the path into a repository, directory, and file
    my ($repo, $dir, $file, @bits) = '', '', '', [];
    while (@path > 0) {
        $repo = join('/', @path);
        # check if remaining path elements refer to a repo
        if (-d "${repository_root}/${repo}/_darcs") {
            if (@bits > 1) {
                $dir  = join('/', @bits[0..$#bits - 1]);
            }
            $file = $bits[$#bits];
            # check if last element of path, stored in $file, is really a dir
            if (-d "${repository_root}/${repo}/_darcs/current/${dir}/${file}") {
                $dir = ($dir ? "$dir/$file" : $file);
                $file = '';
            }
            last;
        } else {
            $repo = '';
            unshift(@bits, pop @path);
        }
    }

    # make sure the repository exists
    unless ($repo) {
        show_error('invalid-repository', '404', 'Invalid repository');
        return;
    }

    # untaint patches and tags. Tags can have arbitrary values, so
    # never pass these unquoted, on pain of pain!
    my $patch = safe_param('p', $hash_regex);
    my $tag   = safe_param('t', '^(.+)$');

    my $darcs_args = "";
    $darcs_args .= " --match 'hash $patch' " if $patch;
    $darcs_args .= " -t '$tag' " if $tag;

    # process the requested command
    if ($cmd eq 'browse') {
        my $fh = dir_listing($repo, $dir);
        transform($fh, $xslt_browse, $xslt_args);
    } elsif ($cmd eq 'patches') {
        my $fh = darcs_xml($repo, "changes", $darcs_args, $dir, $file);
        transform($fh, $xslt_patches, $xslt_args);
    } elsif ($cmd eq 'annotate') {
        $darcs_args .= " --summary --xml-output ";

        my $creator_hash  = safe_param('ch', $hash_regex);
        my $original_path = safe_param('o', $path_regex);

        my $fh = make_xml($repo, $dir, $file);

        # use the creator hash and original file name when available so
        # annotations can span renames
        if ($creator_hash ne '' && $original_path ne '') {
            $darcs_args .= " --creator-hash '$creator_hash' ";
            darcs($fh, $repo, "annotate", $darcs_args, '', $original_path);
        } else {
            darcs($fh, $repo, "annotate", $darcs_args, $dir, $file);
        }

        finish_xml($fh);
        transform($fh, $xslt_annotate, $xslt_args);
    } elsif ($cmd eq 'diff') {
        $darcs_args .= " -u ";
        print "Content-type: text/plain\r\n\r\n";
        darcs(\*STDOUT, $repo, "diff", $darcs_args, $dir, $file);
    } elsif ($cmd eq 'rss') {
        $darcs_args .= " --last 25 ";

        print "Content-type: application/rss+xml\r\n";
        my $fh = darcs_xml($repo, "changes", $darcs_args, $dir, $file);
        transform($fh, $xslt_rss, $xslt_args);
    } else {
        show_error('invalid-command', '400', 'Invalid command');
    }
}

# CGI entry point
respond();
